Index: src/Fl_Pixmap.cxx
===================================================================
--- src/Fl_Pixmap.cxx	(revision 10757)
+++ src/Fl_Pixmap.cxx	(working copy)
@@ -193,24 +193,37 @@
 
 void Fl_Xlib_Graphics_Driver::draw(Fl_Pixmap *pxm, int XP, int YP, int WP, int HP, int cx, int cy) {
   int X, Y, W, H;
+  bool complex_clip = false;
+  Region r = NULL;
   if (pxm->prepare(XP, YP, WP, HP, cx, cy, X, Y, W, H)) return;
   if (pxm->mask_) {
-    // I can't figure out how to combine a mask with existing region,
-    // so cut the image down to a clipped rectangle:
-    int nx, ny; fl_clip_box(X,Y,W,H,nx,ny,W,H);
-    cx += nx-X; X = nx;
-    cy += ny-Y; Y = ny;
     // make X use the bitmap as a mask:
     XSetClipMask(fl_display, fl_gc, pxm->mask_);
-    int ox = X-cx; if (ox < 0) ox += pxm->w();
-    int oy = Y-cy; if (oy < 0) oy += pxm->h();
     XSetClipOrigin(fl_display, fl_gc, X-cx, Y-cy);
+    r = clip_region();
+    if (r && !(XRectInRegion(r, X, Y, W, H) == RectangleIn)) {
+      // At this point, XYWH is the rectangle encompassing the intersection between
+      // the current clip region and the (portion of the) pixmap we have to draw.
+      // The current clip region is often a rectangle. But, when the rounded top angle of a window
+      // is moved above another window, resulting expose events may create a complex clipping region
+      // made of several rectangles. In this case, XYWH contains areas outside the clipping region.
+      complex_clip = true;
+    }
   }
   copy_offscreen(X, Y, W, H, pxm->id_, cx, cy);
   if (pxm->mask_) {
     // put the old clip region back
     XSetClipOrigin(fl_display, fl_gc, 0, 0);
     fl_restore_clip();
+    if (complex_clip) {
+      // At this point, we have just redrawn XYWH of the current window,
+      // and this rectangle contains areas outside the current clipping region which is complex.
+      // We enlarge the current clipping region with XYWH so that any further redraw (e.g., text
+      // drawn over the pixmap) redraws all of XYWH.
+      XRectangle rect = {X,Y,W,H};
+      XUnionRectWithRegion(&rect, r, r); // the current clipping region is modified in place
+      XSetRegion(fl_display, fl_gc, r);  // and declared as new clipping region
+    }
   }
 }
 
